/*
Opis algorytmu accumulate z biblioteki standardowej C++.
Autor: Piotr Kulik
_______________________________________________________________________________

Algorytm: accumulate
Plik nagowkowy: numeric

Opis:
Algorytm wystpuje w dwch wersjach - oglnej i specjalizowanej.
 * Wersja oglna:
	template<class InputIterator, class Type, class BinaryOperation>
	   Type accumulate(
		  InputIterator _First,
		  InputIterator _Last,
		  Type _Val,
		  BinaryOperation _Binary_op
	   );

	- algorytm kumuluje kolekcj elementw w jedn warto liczc
	  kolejne wyniki operacji dwuargumentowej # z argumentami:
	  dotychczas zakumulowanej wartoci i kolejnego elementu
	- cile: wynik powstaje w wyniku operacji
	       wynik = ((((_Val#e1)#e2)#e3)#...)#eN
	  gdzie e1,...,eN jest zakresem elementw wskazywanym przez par iteratorw
	  _First i _Last, # jest operacj dwuargumentow _Binary_op,
	  Val jest wartoci pocztkow

 * Wersja specjalizowana:
	template<class InputIterator, class Type>
	   Type accumulate(
		  InputIterator _First,
		  InputIterator _Last,
		  Type _Val
	   );

	- uywa dodawania (operatora +) jako operacji dwuargumentowej
	- wynikiem dziaania jest suma wszystkich elementw

Uwagi:
- algorytm ma zoono liniow O(n)
- parametr _Binary_op rerezentujcy operator dwuargumentowy moe by obiektem
  funkcyjnym i, jak pokazano w przykadzie 4a, korzysta z rachunku lambda

 */

#include <iostream>
#include <vector>
#include <string>
#include <numeric>
#include <boost/lambda/lambda.hpp>
#include <boost/lambda/bind.hpp>

using namespace std;
using namespace boost::lambda;

/*
 * Prosta klasa figury wykorzystywana w przykadzie 4.
 * Figura ma pole i atrybut okrelajcy widoczno na ekranie.
 */
class Shape {
private:
    float area;
    bool visible;
public:
    Shape() : area(0.0f), visible(true) {}
    Shape(float area, bool visible) : area(area), visible(visible) {}
    void setArea(float area) { this->area = area; }
    float getArea() const { return area; }
    void setVisible(bool visible) { this->visible = visible; }
    bool isVisible() { return visible; }
};

/*
 * Funkcja wykorzystywana w przykadzie 4.
 * Dodaje do zakumulowanej wartoci pole figury, jeli jest ona widoczna.
 */
float sumVisibleArea(float totalArea, Shape shape) {
    if(shape.isVisible())
        return totalArea + shape.getArea();
    return totalArea;
}

int main() {
	/*
	 * Przykady uycia wersji specjalizowanej.
	 * Moemy j wykorzysta dla kadej kolekcji elementw posiadajcych
	 * zdefiniowany operator dodawania.
	 *
	 * Przykad 1. Suma liczb.
	 * Najprostszy moliwy przykad ukazuje sumowanie liczb z wektora i dodanie
	 * do sumy wartoci pocztkowej.
	 */
	vector<int> numbers = vector<int>(5, 10);
	int summed = accumulate(numbers.begin(), numbers.end(), 5);
	cout << summed << endl; // summed = ((((5+10)+10)+10)+10)+10 = 5+5*10 = 55

	/*
	 * Przykad 2. Konkatenacja stringw.
	 * Przykad wykorzystuje przeciony operator + klasy string.
	 */
	vector<string> words;
    words.push_back("To ");
    words.push_back("jest ");
    words.push_back("jedno ");
    words.push_back("zdanie.");
    string concatenated = accumulate(words.begin(), words.end(), string(""));
    cout << concatenated << endl; // concatenated =
                                  // (((""+"To ")+"jest ")+"jedno ")+"zdanie." =
                                  // "To jest jedno zdanie."

	/*
	 * Przykady uycia wersji oglnej.
	 *
	 * Przykad 3. Mnoenie liczb.
	 * Mnoenie liczb z wektora i wartoci pocztkowej.
	 * Moliwe jest wykorzystanie take innych obiektw funkcyjnych
	 * z stl function.
	 */
	vector<int> numbers2 = vector<int>(5, 10);
	int multiplied = accumulate(numbers2.begin(), numbers2.end(), 5, multiplies<int>());
	cout << multiplied << endl; // multiplied = ((((5*10)*10)*10)*10)*10 = 5*10^5 = 50000

	/*
	 * Przykad 4. Pola widocznych figur.
	 * Sumuje pola widocznych figur. Przykad pokazuje uycie bardziej
	 * skomplikowanej funkcji w roli operatora dwuargumentowego.
	 */
	vector<Shape> shapes;
	// drugi parametr konstruktora - czy figura jest widoczna
    Shape s1(12.0f, true);
    Shape s2(5.5f, false);
    Shape s3(10.5f, true);
    shapes.push_back(s1);
    shapes.push_back(s2);
    shapes.push_back(s3);
    float visibleArea = accumulate(shapes.begin(), shapes.end(), 0.0f, sumVisibleArea);
    cout << visibleArea << endl; // visibleArea = (0+12)+10.5 = 22.5

    /*
     * Przykad 4a. Pola figur.
     * Do przystosowywania funkcji (do postaci op. dwuargumentowego)
     * i konstruowania zoonych funkcji mona si take posuy narzdziami
     * rachunku lambda, np. bibliotek boost::bind (wicej o bibliotece na
     * stronie http://www.boost.org/doc/libs/1_37_0/libs/bind/bind.html)
     */
    float totalArea = accumulate (shapes.begin(), shapes.end(), 0.0f,
    		bind(plus<float>(), _1, bind(&Shape::getArea, _2)));
		// totalArea = ((0+12)+5.5)+10.5 = 28
    cout << totalArea << endl;

}
